from typing import Callable, Any, List
from cp93pytools.methodtools import cached_property


class PosetException(Exception):
    message = 'Dummy exception for the poset class'

    def __init__(
        self,
        *hooks: Callable[[], Any],
        message=None,
        args=None,
    ):
        if hooks is not None:
            self.hooks = [*hooks]
        if message is not None:
            self.message = message
        if args is not None:
            self.args = args

    def __str__(self):  # When raised to the top
        for hook in self.hooks:
            hook()
        return repr(self)


# Not poset
class NotPosetException(PosetException):
    message = 'Given relation is not a poset'


class NotTransitiveException(NotPosetException):
    message = 'Given relation is not transitive'


class NotReflexiveException(NotPosetException):
    message = 'Given relation is not reflexive (i<=i)'


class NotAntisymmetricException(NotPosetException):
    message = ('Given relation is not antisymmetric '
               '(antisymmetry: i<=j and j<=i imply j=i)\n'
               'There is a cycle')


class NoBottomsException(NotPosetException):
    message = 'No bottom element found'


class NoTopsException(NotPosetException):
    message = 'No top element found'


# Not lattice
class NotLatticeException(PosetException):
    message = 'Given poset is not a lattice'


class NoLUBException(NotLatticeException):
    message = 'Given poset has pairs with no common upper bounds'


class NotUniqueLUBException(NotLatticeException):
    message = 'Given poset has pairs with non-unique minimal upper bounds'


class NotUniqueBottomException(NotLatticeException):
    message = 'Given poset has multiple bottom elements'


class NotUniqueTopException(NotLatticeException):
    message = 'Given poset has multiple top elements'


class NotCompleteException(PosetException):
    message = 'Given poset is not complete'


class NotDistributiveException(PosetException):
    message = 'Given lattice is not distributive'


class NotModularException(PosetException):
    message = 'Given lattice is not distributive'


class PosetExceptions:
    PosetException = PosetException
    NotPosetException = NotPosetException
    NotTransitiveException = NotTransitiveException
    NotReflexiveException = NotReflexiveException
    NotAntisymmetricException = NotAntisymmetricException
    NotLatticeException = NotLatticeException
    NoLUBException = NoLUBException
    NotUniqueLUBException = NotUniqueLUBException
    NotUniqueBottomException = NotUniqueBottomException
    NotUniqueTopException = NotUniqueTopException
    NotCompleteException = NotCompleteException
    NotDistributiveException = NotDistributiveException
    NoBottomsException = NoBottomsException
    NoTopsException = NoTopsException
    NotModularException = NotModularException



from typing import Any, Callable

from numpy.core.fromnumeric import nonzero


class WBool:
    'Boolean that explains why_false and calls hook when failed to assert'
    value: bool
    why_false: str
    hook: Callable[[], Any]

    def __bool__(self):
        return self.value

    def assert_explain(self):
        if not self:
            try:
                self.hook()
            except:
                pass
            raise AssertionError(self.why_false)

    def __repr__(self):
        return 'True' if self else f'False: {self.why_false}'


class WBools:
    'Class with WBool factory and describe methods'

    def _wbool(self, value: bool, why_false=None):
        'Boolean that describes self when failed to require'
        r = WBool()
        r.value = bool(value)
        r.why_false = why_false
        r.hook = lambda: self.describe()
        return r

    def describe(self):
        pass



import sys


class Outfile:
    '''
    Redirect stdout to a file inside statements like:
    with Outfile(...):
        print(...)
    '''

    def __init__(self, outfile=None):
        self.outfile = outfile

    def __enter__(self):
        if self.outfile is not None:
            self.initial_stdout = sys.stdout
            sys.stdout = open(self.outfile, 'a')

    def __exit__(self, *args):
        if self.outfile is not None:
            sys.stdout.close()
            sys.stdout = self.initial_stdout



from __future__ import annotations
from cp93pytools.methodtools import cached_property
import inspect
import re


class HelpIndex:
    '''
    @section
        Help and examples
    '''

    @classmethod
    def help_index(cls, show_all=False, silent=False):
        # Inspect the source code
        src = inspect.getsource(cls)
        re_sect = r'(?:\n *(@section(?:.|[ \n])+?)(?:\'\'\'|\"\"\"))'
        re_meth = r'(?:def +(.*?\( *self.*?\)):)'
        re_cmeth = r'(?:def +(.*?\( *cls.*?\)):)'
        tokens = re.findall('|'.join((re_meth, re_cmeth, re_sect)), src)

        # Group sections and get docs when available
        sections = []
        methodsOf = {}
        section = '(no section)'
        for f, g, sec in tokens:
            f = f or g
            if f:
                methodsOf[section] = methodsOf.get(section, [])
                name = f[:f.index('(')]
                if not hasattr(cls, name):
                    continue
                func = getattr(cls, name)
                doc = func.__doc__ or ''
                doc = '\n'.join(l.strip() for l in doc.split('\n'))
                if isinstance(func, cached_property):
                    full_f = f'@cached_property\ndef {f}:'
                else:
                    full_f = f'def {f}:'
                methodsOf[section].append((full_f, doc))
            else:
                sections.append((section, methodsOf.get(section, [])))
                section = sec.replace('@section', '').strip()
        sections.append((section, methodsOf.get(section, [])))

        # Write a readable output
        out = []
        for i, (sec, methods) in enumerate(sections):
            out.append(f'\n@section {i}. {sec}\n\n')
            for f, docs in methods:
                underscore = f.startswith(
                    'def _') and not f.startswith('def __')
                if show_all or not underscore or (not underscore and docs):
                    f = '\n'.join(' ' * 4 + s for s in f.split('\n'))
                    docs = '\n'.join(' ' * 8 + s for s in docs.split('\n'))
                    out.append(f'{f}\n')
                    out.append(f'{docs}\n')
                    if docs.strip():
                        out.append('\n')
        out = ''.join(out)

        return (sections, out) if silent else print(out)